/*  $Revision: 1.18 $
**
**  Open the privileged port, then exec innd.
*/
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include "configdata.h"
#include "clibrary.h"
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include "paths.h"
#include "logging.h"
#include "libinn.h"
#include "macros.h"
#if	NOFILE_LIMIT > 0
#if	defined(DO_NEED_TIME)
#include <time.h>
#endif	/* defined(DO_NEED_TIME) */
#include <sys/time.h>
#endif	/* NOFILE_LIMIT > 0 */
#include <sys/resource.h>
#include <errno.h>

/* To include debugging hooks, then set this to the path of your
   debugger. Then set the environment variable INND_DEBUG to a non-empty
   string to. If INND_DEBUG is an absolute path, then that path is passed
   to the debugger instead of the installed location of innd. */
#if 0
#define DEBUGGER "/usr/bin/gdb"
#endif


#if	NOFILE_LIMIT > 0
/*
**  Set the limit on the number of open files we can have.  I don't
**  like having to do this.
*/
STATIC void
SetDescriptorLimit(i)
    int			i;
{
    struct rlimit	rl;

    if (getrlimit(RLIMIT_NOFILE, &rl) < 0) {
	syslog(L_ERROR, "inndstart cant getrlimit(NOFILE) %m");
	return;
    }
    rl.rlim_cur = i;
    if (setrlimit(RLIMIT_NOFILE, &rl) < 0) {
	syslog(L_ERROR, "inndstart cant setrlimit(NOFILE) %d %m", i);
	return;
    }
}
#endif	/* NOFILE_LIMIT > 0 */


int
main(ac, av)
    int			ac;
    char		*av[];
{
    static char		INNDDIR[] = _PATH_INNDDIR;
    GID_T		NewsGID;
    UID_T		NewsUID;
    struct sockaddr_in	server;
    register int	i;
    register int	j;
#if	defined(SO_REUSEADDR)
    int			on;
#endif	/* defined(SO_REUSEADDR) */
    STRING		*argv;
    char		*p;
    char		pflag[SMBUF];
    char		buff[BUFSIZ];
    STRING		env[8];
    struct stat		Sb;
    char 		*dmalenv;

    (void)openlog("inndstart", L_OPENLOG_FLAGS, LOG_INN_PROG);

    /* Make sure INND directory exists. */
    if (stat(INNDDIR, &Sb) < 0 || !S_ISDIR(Sb.st_mode)) {
	syslog(L_FATAL, "inndstart cant stat %s %m", INNDDIR);
	exit(1);
    }
    NewsUID = Sb.st_uid;
    NewsGID = Sb.st_gid;

    if (setgroups(1, &NewsGID) < 0)
        syslog(L_ERROR, "inndstart cant setgroups %m");
    
#if	NOFILE_LIMIT > 0
    SetDescriptorLimit(NOFILE_LIMIT);
#endif	/* NOFILE_LIMIT > 0 */

    /* Create a socket and name it. */
    if ((i = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	syslog(L_FATAL, "inndstart cant socket %m");
	exit(1);
    }
#if	defined(SO_REUSEADDR)
    on = 1;
    if (setsockopt(i, SOL_SOCKET, SO_REUSEADDR, (caddr_t)&on, sizeof on) < 0)
	syslog(L_ERROR, "inndstart cant setsockopt %m");
#endif	/* defined(SO_REUSEADDR) */
    (void)memset((POINTER)&server, 0, sizeof server);
    server.sin_port = htons(NNTP_PORT);
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = htonl(INADDR_ANY);
    if (bind(i, (struct sockaddr *)&server, sizeof server) < 0) {
	syslog(L_FATAL, "inndstart cant bind %m");
	exit(1);
    }
    (void)snprintf(pflag, sizeof(pflag), "-p%d", i);

    /* Build the new argument vector. */
    argv = NEW(STRING, 2 + ac + 1);
    j = 0;
#if defined (DEBUGGER)
    if (getenv ("INND_DEBUG") != NULL) {
    	argv[j++] = DEBUGGER;
	if (*getenv("INND_DEBUG") == '/') {
    		argv[j++] = getenv ("INND_DEBUG") ;
	} else {
    		argv[j++] = _PATH_INND;
	}
    	argv[j] = NULL;
    	(void)printf("Run with -dufp%d\n", i);
    } else {
#endif
        argv[j++] = _PATH_INND;
        argv[j++] = pflag;
        for (i = 1; av[i]; ) {
            if (strncmp(av[i], "-p", 2) != 0)
	        argv[j++] = av[i++];
	    else
	        i++;
        }
        argv[j] = NULL;
#if defined (DEBUGGER)
    }
#endif



    /* 
     * set the initial cpu priority of innd to INND_SETPRIORITY
     * child processes may nice to lower cpu priorites (i.e.,
     * get as much or less cpu than innd)
     */
#if defined(DO_HAVE_SETPRIORITY)
    errno = 0;
    if (setpriority(PRIO_PROCESS, 0, INND_SETPRIORITY) < 0) {
      syslog(L_ERROR, "setpriority failed, errno=%d", errno);
    } else {
      syslog(L_NOTICE, "setpriority: %d", getpriority(PRIO_PROCESS, 0));
    }
#else
    (void)nice(INND_SETPRIORITY);
#endif
    
    /* Set our user and group id. */
    (void)setgid(NewsGID);
    if (getgid() != NewsGID)
	syslog(L_ERROR, "inndstart cant setgid to %d %m", NewsGID);
    (void)setuid(NewsUID);
    if (getuid() != NewsUID)
	syslog(L_ERROR, "inndstart cant setuid to %d %m", NewsUID);

    /* Set up the environment. */
    (void)snprintf(buff, sizeof(buff), "PATH=%s:%s:/bin:/usr/bin:/usr/ucb",
	    _PATH_NEWSBIN, _PATH_NEWSLIB);
    env[0] = COPY(buff);
    (void)snprintf(buff, sizeof(buff), "TMPDIR=%s", _PATH_TMP);
    env[1] = COPY(buff);
    (void)snprintf(buff, sizeof(buff), "SHELL=%s", _PATH_SH);
    env[2] = COPY(buff);
    (void)snprintf(buff, sizeof(buff), "LOGNAME=%s", NEWSMASTER);
    env[3] = COPY(buff);
    (void)snprintf(buff, sizeof(buff), "USER=%s", NEWSMASTER);
    env[4] = COPY(buff);
    (void)snprintf(buff, sizeof(buff), "HOME=%s", _PATH_NEWSHOME);
    env[5] = COPY(buff);
    i = 6;
    /* linux uglyness */
    if ((p = getenv("BIND_INADDR")) != NULL) {
		(void)snprintf(buff, sizeof(buff), "BIND_INADDR=%s", p);
		env[i++] = COPY(buff);
    }
    if ((p = getenv("TZ")) != NULL) {
	(void)snprintf(buff, sizeof(buff), "TZ=%s", p);
	env[i++] = COPY(buff);
    }
#if defined (USE_DMALLOC)
#define DMSTRING "DMALLOC_OPTIONS"
    if ((dmalenv = getenv (DMSTRING)) != NULL) {
       int len = strlen (dmalenv) + 2 + strlen (DMSTRING) ;
       char *q = NEW (char, len) ;
       snprintf (q,len,"%s=%s",DMSTRING,dmalenv) ;
       env[i++] = q ;
    }
#endif
    env[i++] = NULL;

    /* Go exec innd. */
    (void)execve(argv[0], (CSTRING *)argv, (CSTRING *)env);
    syslog(L_FATAL, "inndstart cant exec %s %m", argv[0]);
    _exit(1);
    /* NOTREACHED */
}
